-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("questsTracker", "AceEvent-3.0", "AceHook-3.0")
local _, _, _, interfaceVersion = GetBuildInfo()

-----------------------
--- Library loading ---
-----------------------
local LQTr = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-QuestsTracker")
local LSM = LibStub("LibSharedMedia-3.0")
local LibQTip = LibStub("LibQTip-1.0")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local pairs, unpack = pairs, unpack
local format = string.format
local type = type
local ToggleFrame = ToggleFrame
local SetAbandonQuest, GetAbandonQuestItems, GetAbandonQuestName = SetAbandonQuest, GetAbandonQuestItems, GetAbandonQuestName
local SelectQuestLogEntry = SelectQuestLogEntry
local QuestLogPushQuest = QuestLogPushQuest
local MouseIsOver = MouseIsOver
local ItemRefTooltip = ItemRefTooltip
local GetQuestLogSelection = GetQuestLogSelection
local GetQuestLink = GetQuestLink
local CreateFrame, CreateFont = CreateFrame, CreateFont
local UnitIsDead, UnitAffectingCombat = UnitIsDead, UnitAffectingCombat
local GetSubZoneText, GetZoneText = GetSubZoneText, GetZoneText
local IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown = IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown
local ChatEdit_InsertLink
local ShowUIPanel, HideUIPanel = ShowUIPanel, HideUIPanel

-----------------
--- Variables ---
-----------------
module.modName = LQTr["Quests Tracker"]
local prdb, chdb, options
local dbVersion = 1
local subZone, currentZone, oldZone = "", "", ""
local data = {true, true, true}
local numTrackedQuests = 0
local numTrackedQuestsBML = 0

local tracker, trackerAnchor
local hTrFont = CreateFont("hTrFont")
local qTrFont = CreateFont("qTrFont")
local qoTrFont = CreateFont("qoTrFont")

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		point = "CENTER",
		x = 0,
		y = 0,
		onQuestAccepted = 3,
		onObjectiveProgress = true,
		onSubZone = true,
		onZone = false,
		onDaily = true,
		onClass = true,
		showAnchor = true,
		anchorPosition = 2,
		toggleTracker = true,
		hideTrackerInCombat = true,
		hideCompletedQuest = 3,
		hideCompletedObjective = 3,
		onlyCurrentZone = false,
		limitQuestLevel = false,
		maxLevelDiff = 3,
		questFont = "Friz Quadrata TT",
		questFontSize = 10,
		questObjectiveFont = "Friz Quadrata TT",
		questObjectiveFontSize = 8,
		maxWidth = 300,
		objectivesCellPadding = 5,
		scale = 1,
		backDropColor = {0.09, 0.09, 0.09, 1},
		borderColor = {1, 1, 1, 1},
		colorizeQuestName = true,
		colorizeQuestLevel = true,
		colorizeQuestObjective = true,
		colorizeQuestObjectiveCounter = true,
		displayHeader = true,
		autoCollapseHeader = true,
		headerFont = "Friz Quadrata TT",
		headerFontSize = 12,
		questsCellPadding = 5,
		collapseButtonSize = 24,
		colorizeHeaderName = true,
		hideZoneLevelRange = false,
		colorizeHeaderLevel = true,
		chatLink  = true,
		questsObjectives = true,
		clickTooltip = true,
		shareQuest = true,
		trackQuest = true,
		abandonQuest = true,
		questLog = true,
		overTooltip = true,
		cqiObjectiveStartNPC = true,
		cqiObjectiveEndNPC = true,
		cqiObjective = true,
		mmObjective = true,
	},
}

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("questsTracker", defaults, "Default")
	prdb = self.db.profile
	local dbV = prdb.version or 0
	if dbV < dbVersion then
		for k in pairs(prdb) do
			prdb[k] = nil
		end
	end
	prdb.version = dbVersion
	chdb = self.db.char
	chdb.questsTracked = chdb.questsTracked or {}
	chdb.questsSubZoneAutoTracked = chdb.questsSubZoneAutoTracked or {}
	chdb.questsZoneAutoTracked = chdb.questsZoneAutoTracked or {}
	chdb.zoneDisplay = chdb.zoneDisplay or {}
	hTrFont:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize)
	qTrFont:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize)
	qoTrFont:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize)
end

----------------------------
--- Tracker Anchor Frame ---
----------------------------
local anchor = CreateFrame("Button", nil, UIParent)
anchor:SetHeight(24)
anchor:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = {left = 5, right = 5, top = 5, bottom = 5}, tile = true, tileSize = 16})
anchor:SetBackdropColor(0.09, 0.09, 0.19, 0.5)
anchor:SetBackdropBorderColor(0.5, 0.5, 0.5, 1)
local text = anchor:CreateFontString(nil, nil, "GameFontNormalSmall")
text:SetPoint("CENTER")
text:SetText(LQTr["Tracker"])
anchor:SetWidth(text:GetStringWidth() + 8)
anchor:SetMovable(true)
anchor:RegisterForDrag("LeftButton")
anchor:SetClampedToScreen(true)
anchor:SetScript("OnDragStart", function(self)
	self:StartMoving()
end)
anchor:SetScript("OnDragStop", function(self)
	self:StopMovingOrSizing()
	prdb.point, prdb.x, prdb.y = "BOTTOMLEFT", self:GetLeft(), self:GetBottom()
	trackerAnchor = anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	LibQTip:Release(tracker)
	module:updateQuestTracker()
end)

---------------
--- Options ---
---------------
function module:getOptions()
	if not options then
		options = {
			type = "group",
			name = module.modName,
			desc = module.modName,
			get = function(info) 
				if type(prdb[info[#info]]) == "table" then
					return unpack(prdb[info[#info]])
				else
					return prdb[info[#info]]
				end
			end,
			set = function(info, arg1, arg2, arg3, arg4)
				if arg2 then
					prdb[info[#info]][1] = arg1
					prdb[info[#info]][2] = arg2
					prdb[info[#info]][3] = arg3
					prdb[info[#info]][4] = arg4
				else
					prdb[info[#info]] = arg1
					if info[#info] == "showAnchor" then
						if prdb.showAnchor then
							anchor:Show()
						else
							anchor:Hide()
						end
					elseif info[#info] == "hideTrackerInCombat" then
						if prdb.hideTrackerInCombat then
							module:RegisterEvent("PLAYER_REGEN_ENABLED","updateQuestsTracker")
							module:RegisterEvent("PLAYER_REGEN_DISABLED","updateQuestsTracker")
						else
							module:UnregisterEvent("PLAYER_REGEN_ENABLED")
							module:UnregisterEvent("PLAYER_REGEN_DISABLED")
						end
					elseif info[#info] == "headerFont" or info[#info] == "headerFontSize" then
						hTrFont:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize)
					elseif info[#info] == "questFont" or info[#info] == "questFontSize" then
						qTrFont:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize)
					elseif info[#info] == "questObjectiveFont" or info[#info] == "questObjectiveFontSize" then
						qoTrFont:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize)
					elseif info[#info] == "maxLevelDiff" then
						core:updateMaxLevelDiff(value)
						module:questsWeightedLevelUpdated()
					end
				end
				LibQTip:Release(tracker)
				module:updateQuestTracker()
			end,
			args = {
				onQuestAccepted = {
					name = LQTr["Auto track quests when you accept them"],
					type = "select",
					values = {[1] = LQTr["Always"], [2] = LQTr["Below Max Level"], [3] = LQTr["Never"]},
					width = "full",
					order = 0,
				},
				onObjectiveProgress = {
					name = LQTr["Auto track quests on objective progress"],
					type = "toggle",
					width = "full",
					order = 1,
				},
				onSubZone = {
					name = LQTr["Auto track quests on sub zone change"],
					type = "toggle",
					width = "full",
					order = 2,
				},
				onZone = {
					name = LQTr["Auto track quests on zone change"],
					type = "toggle",
					width = "full",
					order = 3,
				},
				onDaily = {
					name = LQTr["Auto track daily quests"],
					type = "toggle",
					width = "full",
					order = 4,
				},
				onClass = {
					name = LQTr["Auto track class quests"],
					type = "toggle",
					width = "full",
					order = 5,
				},
				showAnchor = {
					name = LQTr["Show quest tracker anchor"],
					type = "toggle",
					width = "full",
					order = 6,
				},
				anchorPosition = {
					name = LQTr["Anchor position"],
					type = "select",
					values = {[1] = "TOPLEFT", [2] = "TOP", [3] = "TOPRIGHT", [4] = "BOTTOMLEFT", [5] = "BOTTOM", [6] = "BOTTOMRIGHT"},
					width = "full",
					order = 7,
				},
				trackerHeader = {
					type = "header",
					name = LQTr["Tracker Settings"],
					order = 8,
				},
				displayHeader = {
					name = LQTr["Display quests by zone"],
					type = "toggle",
					width = "full",
					order = 9,
				},
				autoCollapseHeader = {
					name = LQTr["Automatically expand /collapse zone header on zone change"],
					hidden = function()
						if prdb.displayHeader then
							return false
						else
							return true
						end
					end,
					type = "toggle",
					width = "full",
					order = 10,
				},
				toggleTracker = {
					name = LQTr["Display tracker"],
					type = "toggle",
					width = "full",
					order = 11,
				},
				hideTrackerInCombat = {
					name = LQTr["Hide quest tracker in combat"],
					type = "toggle",
					width = "full",
					order = 12,
				},
				hideCompletedQuest = {
					name = LQTr["Hide completed quest"],
					type = "select",
					values = {[1] = LQTr["When you've completed the quest"], [2] = LQTr["When everyone in the party have completed the quest"], [3] = LQTr["Never"]},
					width = "full",
					order = 13,
				},
				hideCompletedObjective = {
					name = LQTr["Hide completed quests Objectives"],
					type = "select",
					values = {[1] = LQTr["When you've completed the objective"], [2] = LQTr["When everyone in the party have completed the objective"], [3] = LQTr["Never"]},
					width = "full",
					order = 14,
				},
				onlyCurrentZone = {
					name = LQTr["Hide quests that are not in the current zone"],
					type = "toggle",
					width = "full",
					order = 15,
				},
				limitQuestLevel = {
					name = LQTr["Hide high level quests (You'll have to choose the maximum level difference)"],
					type = "toggle",
					width = "full",
					order = 16,
				},
				maxLevelDiff = {
					name = LQTr["Maximum level difference between the weighted quest level and your level"],
					hidden = function()
						if prdb.limitQuestLevel then
							return false
						else
							return true
						end
					end,
					type = "range",
					min = 0,
					max = 10,
					step = 1,
					width = "full",
					order = 17,
				},
				headerFont = {
					name = LQTr["Zone header font"],
					hidden = function()
						if prdb.displayHeader then
							return false
						else
							return true
						end
					end,
					type = "select",
					dialogControl = 'LSM30_Font',
					values = AceGUIWidgetLSMlists.font,
					width = "full",
					order = 18,
				},
				headerFontSize = {
					name = LQTr["Zone header font size"],
					hidden = function()
						if prdb.displayHeader then
							return false
						else
							return true
						end
					end,
					type = "range",
					min = 6,
					max = 20,
					step = 1,
					width = "full",
					order = 19,
				},
				questFont = {
					name = LQTr["Quests font"],
					type = "select",
					dialogControl = 'LSM30_Font',
					values = AceGUIWidgetLSMlists.font,
					width = "full",
					order = 20,
				},
				questFontSize = {
					name = LQTr["Quests font size"],
					type = "range",
					min = 6,
					max = 20,
					step = 1,
					width = "full",
					order = 21,
				},
				questObjectiveFont = {
					name = LQTr["Quests objectives font"],
					type = "select",
					dialogControl = 'LSM30_Font',
					values = AceGUIWidgetLSMlists.font,
					width = "full",
					order = 22,
				},
				questObjectiveFontSize = { 
					name = LQTr["Quests objectives font size"],
					type = "range",
					min = 6,
					max = 20,
					step = 1,
					width = "full",
					order = 23,
				},
				maxWidth = {
					name = LQTr["Quests Tracker column max width"],
					type = "range",
					min = 50,
					max = 600,
					step = 5,
					width = "full",
					order = 24,
				},
				scale = {
					name = LQTr["Quests Tracker scale"],
					type = "range",
					min = 0,
					max = 1,
					step = 0.1,
					width = "full",
					order = 25,
				},
				questsCellPadding = {
					name = LQTr["Quests cell padding"],
					hidden = function()
						if prdb.displayHeader then
							return false
						else
							return true
						end
					end,
					type = "range",
					min = 0,
					max = 50,
					step = 1,
					width = "full",
					order = 26,
				},
				objectivesCellPadding = {
					name = LQTr["Quests objectives cell padding"],
					type = "range",
					min = 0,
					max = 50,
					step = 1,
					width = "full",
					order = 27,
				},
				collapseButtonSize = {
					name = LQTr["Collapse Button Size"],
					hidden = function()
						if prdb.displayHeader then
							return false
						else
							return true
						end
					end,
					type = "range",
					min = 10,
					max = 40,
					step = 1,
					width = "full",
					order = 28,
				},
				backDropColor = {
					name = LQTr["Quests Tracker background color"],
					type = "color",
					hasAlpha = true,
					width = "full",
					order = 29,
				},
				borderColor = {
					name = LQTr["Quests Tracker border color"],
					type = "color",
					hasAlpha = true,
					width = "full",
					order = 30,
				},
				colorizeHeaderName = {
					name = LQTr["Colorize Quests Header Name"],
					hidden = function()
						if prdb.displayHeader then
							return false
						else
							return true
						end
					end,
					type = "toggle",
					width = "full",
					order = 31,
				},
				hideZoneLevelRange = {
					name = LQTr["Hide zone level range"],
					hidden = function()
						if prdb.displayHeader then
							return false
						else
							return true
						end
					end,
					type = "toggle",
					width = "full",
					order = 32,
				},
				colorizeHeaderLevel = {
					name = LQTr["Colorize Quests Header Level"],
					hidden = function()
						if prdb.displayHeader and not prdb.hideZoneLevelRange then
							return false
						else
							return true
						end
					end,
					type = "toggle",
					width = "full",
					order = 33,
				},
				colorizeQuestName = {
					name = LQTr["Colorize Quests Name"],
					type = "toggle",
					width = "full",
					order = 34,
				},
				colorizeQuestLevel = {
					name = LQTr["Colorize Quests Level"],
					type = "toggle",
					width = "full",
					order = 35,
				},
				colorizeQuestObjective = {
					name = LQTr["Colorize Quests Objectives"],
					type = "toggle",
					width = "full",
					order = 36,
				},
				colorizeQuestObjectiveCounter = {
					name = LQTr["Colorize quests objectives counter"],
					type = "toggle",
					width = "full",
					order = 37,
				},
				trackerInteraction = {
					type = "header",
					name = LQTr["Tracker interaction"],
					order = 38,
				},
				chatLink  = {
					name = LQTr["Insert quest link in chat on shift right click"],
					type = "toggle",
					width = "full",
					order = 39,
				},
				questsObjectives = {
					name = LQTr["Insert quest objectives in chat on ctrl right click"],
					type = "toggle",
					width = "full",
					order = 40,
				},
				clickTooltip = {
					name = LQTr["Display of quest tooltip on right click"],
					type = "toggle",
					width = "full",
					order = 41,
				},
				shareQuest = {
					name = LQTr["Share quest on ctrl left click"],
					type = "toggle",
					width = "full",
					order = 42,
				},
				trackQuest = {
					name = LQTr["Track quest on shift left click"],
					type = "toggle",
					width = "full",
					order = 43,
				},
				abandonQuest = {
					name = LQTr["Abandon quest on alt left click"],
					type = "toggle",
					width = "full",
					order = 44,
				},
				questLog = {
					name = LQTr["Display quest log on left click"],
					type = "toggle",
					width = "full",
					order = 45,
				},
				overTooltip = {
					name = LQTr["Display quest tooltip on mouse over"],
					type = "toggle",
					width = "full",
					order = 46,
				},
				cqiObjectiveStartNPC = {
					name = LQTr["Display of start npc in Cartographer quest Info"],
					hidden = function()
						if Cartographer_QuestInfo then
							return false
						else
							return true
						end
					end,
					type = "toggle",
					width = "full",
					order = 47,
				},
				cqiObjectiveEndNPC = {
					name = LQTr["Toggle display of end npc in Cartographer Quest Info"],
					hidden = function()
						if Cartographer_QuestInfo then
							return false
						else
							return true
						end
					end,
					type = "toggle",
					width = "full",
					order = 48,
				},
				cqiObjective = {
					name = LQTr["Display of objective in Cartographer Quest Info"],
					hidden = function()
						if Cartographer_QuestInfo then
							return false
						else
							return true
						end
					end,
					type = "toggle",
					width = "full",
					order = 49,
				},
				mmObjective = {
					name = LQTr["Display of objective in your map with MobMap"],
					hidden = function()
						if MobMapDotParentFrame then
							return false
						else
							return true
						end
					end,
					type = "toggle",
					width = "full",
					order = 50,
				},
			},
		}
	end
	return options
end

function module:OnEnable()
	subZone = GetSubZoneText() or ""
	currentZone = GetZoneText() or ""
	anchor:SetPoint(prdb.point, prdb.x, prdb.y)
	if not prdb.showAnchor then anchor:Hide() else anchor:Show() end
	trackerAnchor = anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	if prdb.hideTrackerInCombat then
		self:RegisterEvent("PLAYER_REGEN_ENABLED","updateQuestsTracker")
		self:RegisterEvent("PLAYER_REGEN_DISABLED","updateQuestsTracker")
	end
	self:RegisterEvent("MINIMAP_ZONE_CHANGED", "zoneChanged")
	self:RegisterEvent("ZONE_CHANGED", "zoneChanged")
	self:RegisterEvent("ZONE_CHANGED_INDOORS", "zoneChanged")
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA", "zoneChanged")
	self:SecureHook("AddQuestWatch")
	self:SecureHook("RemoveQuestWatch")
	self:SecureHook("GetNumQuestWatches")
	if interfaceVersion == 30100 then
		self:SecureHook("WatchFrame_Update")
		WATCHFRAME_MAXQUESTS = 100
		WATCHFRAME_MAXACHIEVEMENTS = 100
		WATCHFRAME_LINEHEIGHT = 1
	else
		self:SecureHook("QuestWatch_Update")
		MAX_WATCHABLE_QUESTS = MAX_QUESTS
	end
	core.RegisterCallback(self, "toggleTracker")
	core.RegisterCallback(self, "endPlayerQuest")
	core.RegisterCallback(self, "endPartyQuest")
	core.RegisterCallback(self, "playerQuestComplete", "playerQuestChanged")
	core.RegisterCallback(self, "playerQuestFailed", "playerQuestChanged")
	core.RegisterCallback(self, "partyQuestComplete", "partyQuestChanged")
	core.RegisterCallback(self, "partyQuestFailed", "partyQuestChanged")
	core.RegisterCallback(self, "playerQuestObjectiveProgress")
	core.RegisterCallback(self, "partyQuestObjectiveProgress")
	core.RegisterCallback(self, "playerQuestLost")
	core.RegisterCallback(self, "partyQuestLost")
	core.RegisterCallback(self, "toggleTrackedQuest")
	core.RegisterCallback(self, "questObjectiveAdded")
	core.RegisterCallback(self, "questDescriptionAdded")
	core.RegisterCallback(self, "updateQuestsTracker", "updateQuestsTracker")
	core.RegisterCallback(self, "questsWeightedLevelUpdated")
end

function module:OnDisable()
	anchor:Hide()
	LibQTip:Release(tracker)
	tracker = nil
	core.UnregisterAllCallbacks(self)
end

--------------
--- Events ---
--------------
function module:zoneChanged()
	if UnitIsDead("player") then return end
	local trackerNeedUpdate = false
	subZone = GetSubZoneText() or ""
	currentZone = GetZoneText() or ""
	if prdb.onZone and currentZone ~= oldZone then
		for uid in pairs(chdb.questsZoneAutoTracked) do
			if core.quests[uid] then
				trackerNeedUpdate = true
				module:RemoveQuestWatch(nil, uid, true)
			else
				chdb.questsZoneAutoTracked[uid] = nil
			end
		end
		for uid, qData in pairs(core.quests) do
			if not chdb.questsTracked[uid] and qData.zone == currentZone then
				chdb.questsZoneAutoTracked[uid] = true
				trackerNeedUpdate = true
				module:AddQuestWatch(nil, nil, uid, true)
			end
		end
	end
	if prdb.onSubZone and subZone ~= "" then
		for uid in pairs(chdb.questsSubZoneAutoTracked) do
			if core.quests[uid] then
				trackerNeedUpdate = true
				module:RemoveQuestWatch(nil, uid, true)
			else
				chdb.questsSubZoneAutoTracked[uid] = nil
			end
		end
		for uid, qData in pairs(core.quests) do
			if not chdb.questsTracked[uid] and ((qData.objective and qData.objective:find(subZone, 1, true)) or (qData.description and qData.description:find(subZone, 1, true))) then
				chdb.questsSubZoneAutoTracked[uid] = true
				trackerNeedUpdate = true
				module:AddQuestWatch(nil, nil, uid, true)
			end
		end
	end
	if prdb.displayHeader and prdb.autoCollapseHeader then
		for zone, zData in pairs(core.questsByZone) do
			if currentZone == zone then
				chdb.zoneDisplay[zone] = 1
			elseif zData.zone then
				chdb.zoneDisplay[zone] = 0
			else
				chdb.zoneDisplay[zone] = 1
			end
		end
		trackerNeedUpdate = true
	end
	if trackerNeedUpdate then
		LibQTip:Release(tracker)
		module:updateQuestTracker()
	end
	oldZone = currentZone
end
-------------
--- Hooks ---
-------------
function module:AddQuestWatch(id, timeToTrack, uid, batch)
	if not uid then
		uid = core.questsIdToUid[id]
		if not uid then return end
	end
	if (not chdb.questsTracked[uid]) or (chdb.questsTracked[uid] and core.quests[uid].tracked == 0) then
		numTrackedQuests = numTrackedQuests + 1
		core.numPlayerCompletedTrackedQuests = core.numPlayerCompletedTrackedQuests + core.quests[uid].playerComplete
		core.numPartyCompletedTrackedQuests = core.numPartyCompletedTrackedQuests + core.quests[uid].partyComplete
		chdb.questsTracked[uid] = true
		core.quests[uid].tracked = 1
		if core.quests[uid].complete[core.playerName] then
			core.questsByZone[core.quests[uid].zone].numTrackedQuests = core.questsByZone[core.quests[uid].zone].numTrackedQuests + 1
			core.questsByZone[core.quests[uid].zone].playerQuestsComplete = core.questsByZone[core.quests[uid].zone].playerQuestsComplete + core.quests[uid].playerComplete
			core.questsByZone[core.quests[uid].zone].partyQuestsComplete = core.questsByZone[core.quests[uid].zone].partyQuestsComplete + core.quests[uid].partyComplete
			if (core.quests[uid].weightedLevel - core.playerLevel <= prdb.maxLevelDiff) then
				core.questsByZone[core.quests[uid].zone].questsBML = core.questsByZone[core.quests[uid].zone].questsBML + 1
				core.questsByZone[core.quests[uid].zone].playerQuestsCompleteBML = core.questsByZone[core.quests[uid].zone].playerQuestsCompleteBML + core.quests[uid].playerComplete
				core.questsByZone[core.quests[uid].zone].partyQuestsCompleteBML = core.questsByZone[core.quests[uid].zone].partyQuestsCompleteBML + core.quests[uid].partyComplete
				core.numPlayerCompletedTrackedQuestsBML = core.numPlayerCompletedTrackedQuestsBML + 1
				core.numPartyCompletedTrackedQuestsBML = core.numPartyCompletedTrackedQuestsBML + 1
			end
		else
			core.questsByZone[LQTr["PARTYQUESTSZONE"]].numTrackedQuests = core.questsByZone[LQTr["PARTYQUESTSZONE"]].numTrackedQuests + 1
			core.questsByZone[LQTr["PARTYQUESTSZONE"]].partyQuestsComplete = core.questsByZone[LQTr["PARTYQUESTSZONE"]].partyQuestsComplete + core.quests[uid].partyComplete
			if (core.quests[uid].weightedLevel - core.playerLevel <= prdb.maxLevelDiff) then
				core.questsByZone[LC["PARTYQUESTSZONE"]].questsBML = core.questsByZone[LC["PARTYQUESTSZONE"]].questsBML + 1
				core.questsByZone[LC["PARTYQUESTSZONE"]].partyQuestsCompleteBML = core.questsByZone[LC["PARTYQUESTSZONE"]].partyQuestsCompleteBML + core.quests[uid].partyComplete
				core.numPlayerCompletedTrackedQuestsBML = core.numPlayerCompletedTrackedQuestsBML + 1
				core.numPartyCompletedTrackedQuestsBML = core.numPartyCompletedTrackedQuestsBML + 1
			end
		end
		if not batch then
			LibQTip:Release(tracker)
			module:updateQuestTracker()
		end
	end
end

function module:RemoveQuestWatch(id, uid, batch)
	if not uid then
		uid = core.questsIdToUid[id]
		if not uid then return end
	end
	if chdb.questsTracked[uid] then
		numTrackedQuests = numTrackedQuests - 1
		core.numPlayerCompletedTrackedQuests = core.numPlayerCompletedTrackedQuests - core.quests[uid].playerComplete
		core.numPartyCompletedTrackedQuests = core.numPartyCompletedTrackedQuests - core.quests[uid].partyComplete
		chdb.questsTracked[uid] = nil
		chdb.questsSubZoneAutoTracked[uid] = nil
		chdb.questsZoneAutoTracked[uid] = nil
		core.quests[uid].tracked = 0
		if core.quests[uid].complete[core.playerName] then
			core.questsByZone[core.quests[uid].zone].numTrackedQuests = core.questsByZone[core.quests[uid].zone].numTrackedQuests - 1
			core.questsByZone[core.quests[uid].zone].playerQuestsComplete = core.questsByZone[core.quests[uid].zone].playerQuestsComplete - core.quests[uid].playerComplete
			core.questsByZone[core.quests[uid].zone].partyQuestsComplete = core.questsByZone[core.quests[uid].zone].partyQuestsComplete - core.quests[uid].partyComplete
			if (core.quests[uid].weightedLevel - core.playerLevel <= prdb.maxLevelDiff) then
				core.questsByZone[core.quests[uid].zone].questsBML = core.questsByZone[core.quests[uid].zone].questsBML - 1
				core.questsByZone[core.quests[uid].zone].playerQuestsCompleteBML = core.questsByZone[core.quests[uid].zone].playerQuestsCompleteBML - core.quests[uid].playerComplete
				core.questsByZone[core.quests[uid].zone].partyQuestsCompleteBML = core.questsByZone[core.quests[uid].zone].partyQuestsCompleteBML - core.quests[uid].partyComplete
				core.numPlayerCompletedTrackedQuestsBML = core.numPlayerCompletedTrackedQuestsBML - 1
				core.numPartyCompletedTrackedQuestsBML = core.numPartyCompletedTrackedQuestsBML - 1
			end
		else
			core.questsByZone[LQTr["PARTYQUESTSZONE"]].numTrackedQuests = core.questsByZone[LQTr["PARTYQUESTSZONE"]].numTrackedQuests - 1
			core.questsByZone[LQTr["PARTYQUESTSZONE"]].partyQuestsComplete = core.questsByZone[LQTr["PARTYQUESTSZONE"]].partyQuestsComplete - core.quests[uid].partyComplete
			if (core.quests[uid].weightedLevel - core.playerLevel <= prdb.maxLevelDiff) then
				core.questsByZone[LC["PARTYQUESTSZONE"]].questsBML = core.questsByZone[LC["PARTYQUESTSZONE"]].questsBML - 1
				core.questsByZone[LC["PARTYQUESTSZONE"]].partyQuestsCompleteBML = core.questsByZone[LC["PARTYQUESTSZONE"]].partyQuestsCompleteBML - core.quests[uid].partyComplete
				core.numPlayerCompletedTrackedQuestsBML = core.numPlayerCompletedTrackedQuestsBML - 1
				core.numPartyCompletedTrackedQuestsBML = core.numPartyCompletedTrackedQuestsBML - 1
			end
		end
		if not batch then
			LibQTip:Release(tracker)
			module:updateQuestTracker()
		end
	end
end

function module:GetNumQuestWatches()
	return numTrackedQuests
end

function module:QuestWatch_Update()
	QuestWatchFrame:Hide()
end

function module:WatchFrame_Update()
	WatchFrame:Hide()
	WatchFrameTitleButton:Hide()
	for _, v in pairs(WATCHFRAME_TIMERLINES) do
		v:Hide()
	end
	for _, v in pairs(WATCHFRAME_ACHIEVEMENTLINES) do
		v:Hide()
	end
	for _, v in pairs(WATCHFRAME_QUESTLINES) do
		v:Hide()
	end
	for _, v in pairs(WATCHFRAME_LINKBUTTONS) do
		v:Hide()
	end
end

function module:isQuestTracked(uid)
	if chdb.questsTracked[uid] then
		return true
	else
		return false
	end
end

-----------------
--- Callbacks ---
-----------------
function module:questsWeightedLevelUpdated(event)
	for zone, zData in pairs(core.questsByZone) do
		core.questsByZone[zone].questsBML = 0
		core.questsByZone[zone].playerQuestsCompleteBML = 0
		core.questsByZone[zone].partyQuestsCompleteBML = 0
		for uid in pairs(zData.quests) do
 			if  core.quests[uid].tracked == 1 then --and (core.quests[uid].weightedLevel - core.playerLevel <= prdb.maxLevelDiff) then
 				core.questsByZone[zone].questsBML = core.questsByZone[zone].questsBML + 1
 				core.questsByZone[zone].playerQuestsCompleteBML = core.questsByZone[zone].playerQuestsCompleteBML + core.quests[uid].playerComplete
 				core.questsByZone[zone].partyQuestsCompleteBML = core.questsByZone[zone].partyQuestsCompleteBML + core.quests[uid].partyComplete
 			end
		end
	end
end

function module:toggleTracker(event)
	prdb.toggleTracker = not prdb.toggleTracker
	LibQTip:Release(tracker)
	module:updateQuestTracker()
end

function module:endPlayerQuest(event, first, commHandler, uid, started)
	local questAddedToTracker = 0
	if first then
		if chdb.questsTracked[uid] and core.quests[uid].tracked == 0 then
			AddQuestWatch(core.quests[uid].id, nil, uid, first)
		end
	elseif prdb.onQuestAccepted == 1 or (prdb.onQuestAccepted == 2 and core.playerLevel ~= MAX_PLAYER_LEVEL) then
		if not chdb.questsTracked[uid] then
			questAddedToTracker = 1
			AddQuestWatch(core.quests[uid].id, nil, uid, first)
		else
			chdb.questsSubZoneAutoTracked[uid] = nil
			chdb.questsZoneAutoTracked[uid] = nil
		end
	else
		if not chdb.questsTracked[uid] then
			if (prdb.onObjectiveProgress and started == 1) or (prdb.onDaily and core.quests[uid].daily == DAILY) or (prdb.onClass and core.quests[uid].class) or (prdb.onZone and core.quests[uid].zone == currentZone) then
				questAddedToTracker = 1
				AddQuestWatch(core.quests[uid].id, nil, uid, first)
			elseif prdb.onSubZone and subZone ~= "" and (core.quests[uid].objective:find(subZone, 1, true) or core.quests[uid].description:find(subZone, 1, true)) then
				chdb.questsSubZoneAutoTracked[uid] = true
				questAddedToTracker = 1
				AddQuestWatch(core.quests[uid].id, nil, uid, first)
			end
		elseif (prdb.onObjectiveProgress and started == 1) or (prdb.onDaily and core.quests[uid].tag == DAILY) or (prdb.onClass and core.quests[uid].class) or (prdb.onZone and core.quests[uid].zone == currentZone) then
			chdb.questsSubZoneAutoTracked[uid] = nil
			chdb.questsZoneAutoTracked[uid] = nil
		end
	end
	if not first and questAddedToTracker == 0 and chdb.questsTracked[uid] then
		LibQTip:Release(tracker)
		module:updateQuestTracker()
	end
end

function module:endPartyQuest(event, sender, uid, started, batch)
	local questAddedToTracker = 0
	if batch then
		if chdb.questsTracked[uid] and core.quests[uid].tracked == 0 then
			AddQuestWatch(core.quests[uid].id, nil, uid, batch)
		end
	elseif prdb.onQuestAccepted == 1 or (prdb.onQuestAccepted == 2 and core.playerLevel ~= MAX_PLAYER_LEVEL) then
		if not chdb.questsTracked[uid] then
			questAddedToTracker = 1
			AddQuestWatch(core.quests[uid].id, nil, uid, batch)
		else
			chdb.questsSubZoneAutoTracked[uid] = nil
			chdb.questsZoneAutoTracked[uid] = nil
		end
	else
		if not chdb.questsTracked[uid] and (prdb.onObjectiveProgress and started == 1) or (prdb.onDaily and core.quests[uid].tag == DAILY) then
			questAddedToTracker = 1
			AddQuestWatch(core.quests[uid].id, nil, uid, batch)
		elseif (prdb.onObjectiveProgress and started == 1) or (prdb.onDaily and core.quests[uid].tag == DAILY) then
			chdb.questsSubZoneAutoTracked[uid] = nil
			chdb.questsZoneAutoTracked[uid] = nil
		end
	end
	if not batch and questAddedToTracker == 0 and chdb.questsTracked[uid] then

		LibQTip:Release(tracker)
		module:updateQuestTracker()
	end
end

function module:playerQuestChanged(event, first, commHandler, uid)
	if chdb.questsTracked[uid] then
		LibQTip:Release(tracker)
		module:updateQuestTracker()
	end
end

function module:partyQuestChanged(event, sender, uid)
	if chdb.questsTracked[uid] then
		LibQTip:Release(tracker)
		module:updateQuestTracker()
	end
end

function module:playerQuestObjectiveProgress(event, first, commHandler, uid, objective, got, nGot)
	if prdb.onObjectiveProgress then
		chdb.questsSubZoneAutoTracked[uid] = nil
		chdb.questsZoneAutoTracked[uid] = nil
		if not chdb.questsTracked[uid] then
			AddQuestWatch(core.quests[uid].id, nil, uid, nil)
		else
			LibQTip:Release(tracker)
			module:updateQuestTracker()
		end
	elseif chdb.questsTracked[uid] then
		LibQTip:Release(tracker)
		module:updateQuestTracker()
	end
end

function module:partyQuestObjectiveProgress(event, sender, uid, objective, got, nGot)
	if prdb.onObjectiveProgress then
		chdb.questsSubZoneAutoTracked[uid] = nil
		chdb.questsZoneAutoTracked[uid] = nil
		if not chdb.questsTracked[uid] then
			AddQuestWatch(core.quests[uid].id, nil, uid, nil)
		else
			LibQTip:Release(tracker)
			module:updateQuestTracker()
		end
	elseif chdb.questsTracked[uid] then
		LibQTip:Release(tracker)
		module:updateQuestTracker()
	end
end

function module:playerQuestLost(event, first, commHandler, uid)
	if not core.quests[uid] and chdb.questsTracked[uid] then
		chdb.questsTracked[uid] = nil
		chdb.questsSubZoneAutoTracked[uid] = nil
		chdb.questsZoneAutoTracked[uid] = nil
		numTrackedQuests = numTrackedQuests - 1
		LibQTip:Release(tracker)
		module:updateQuestTracker()
	elseif core.quests[uid] and chdb.questsTracked[uid] then
		LibQTip:Release(tracker)
		module:updateQuestTracker()
	end
end

function module:partyQuestLost(event, sender, uid, title)
	if not core.quests[uid] and chdb.questsTracked[uid] then
		chdb.questsTracked[uid] = nil
		chdb.questsSubZoneAutoTracked[uid] = nil
		chdb.questsZoneAutoTracked[uid] = nil
		numTrackedQuests = numTrackedQuests - 1
		LibQTip:Release(tracker)
		module:updateQuestTracker()
	elseif core.quests[uid] and chdb.questsTracked[uid] then
		LibQTip:Release(tracker)
		module:updateQuestTracker()
	end
end

function module:toggleTrackedQuest(event, uid)
	if chdb.questsTracked[uid] then
		RemoveQuestWatch(core.quests[uid].id, uid, nil)
	elseif not chdb.questsTracked[uid] then
		AddQuestWatch(core.quests[uid].id, nil, uid, nil)
	end
end

function module:questObjectiveAdded(event, uid)
	if prdb.onSubZone and subZone ~= "" then
		if not chdb.questsTracked[uid] and core.quests[uid].objective and core.quests[uid].objective:find(subZone, 1, true) then
			chdb.questsSubZoneAutoTracked[uid] = true
			AddQuestWatch(core.quests[uid].id, nil, uid, nil)
		elseif chdb.questsTracked[uid] then
			LibQTip:Release(tracker)
			module:updateQuestTracker()
		end
	elseif chdb.questsTracked[uid] then
		LibQTip:Release(tracker)
		module:updateQuestTracker()
	end
end

function module:questDescriptionAdded(event, uid)
	if prdb.onSubZone and subZone ~= "" then
		if not chdb.questsTracked[uid] and core.quests[uid].description	and core.quests[uid].description:find(subZone, 1, true) then
			chdb.questsSubZoneAutoTracked[uid] = true
			AddQuestWatch(core.quests[uid].id, nil, uid, nil)
		elseif chdb.questsTracked[uid] then
			LibQTip:Release(tracker)
			module:updateQuestTracker()
		end
	elseif chdb.questsTracked[uid] then
		LibQTip:Release(tracker)
		module:updateQuestTracker()
	end
end

function module:updateQuestsTracker(event)
	LibQTip:Release(tracker)
	module:updateQuestTracker()
end

----------------------------
--- Custom Cell Provider ---
----------------------------
local cProvider, cPrototype, basecCellPrototype = LibQTip:CreateCellProvider(LibQTip.LabelProvider)
function cPrototype:InitializeCell()
	self.fontString = self:CreateFontString()
	self.fontString:SetFontObject(GameTooltipText)
end

function cPrototype:SetupCell(tooltip, value, justification, font, bType, bValue, bValue2, maxWidth, indent, anchor)
	self:EnableMouse(true)
	self:SetScript("OnMouseDown", function(frame, button)
		if bType == "zone" then
			if chdb.zoneDisplay[bValue] == 1 then
				chdb.zoneDisplay[bValue] = 0
			else
				chdb.zoneDisplay[bValue] = 1
			end
			tracker:SetScript("OnLeave", nil)
			LibQTip:Release(tracker)
			tracker = nil
			module:updateQuestTracker()
		elseif bType == "quest" then
			if button == "RightButton" then
				if IsShiftKeyDown() then
					if core.quests[bValue].complete[core.playerName] and core.quests[bValue].id and prdb.chatLink then
						ChatFrameEditBox:Show()
						ChatEdit_InsertLink(GetQuestLink(core.quests[bValue].id))
					end
				elseif IsControlKeyDown() then
					if core.quests[bValue].numObjectives > 0 and core.quests[bValue].complete[core.playerName] and prdb.questsObjectives then
						ChatFrameEditBox:Show()
						for objective, qoData in pairs(core.questsObjectives[bValue]) do
							local got, _ = qoData:match("^(.-)~(%d-)$")
							ChatFrameEditBox:Insert(format("{%s %s/%s} ", qoData.desc, got, qoData.need))
						end
					end
				elseif core.quests[bValue].complete[core.playerName] and core.quests[bValue].id and prdb.clickTooltip then
					ItemRefTooltip:ClearLines()
					ShowUIPanel(ItemRefTooltip)
					if not ItemRefTooltip:IsVisible() then
						ItemRefTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
					end
					local link = GetQuestLink(core.quests[bValue].id)
					if link then
						ItemRefTooltip:SetHyperlink(link)
					end
				end
			elseif button == "LeftButton" then
				if IsControlKeyDown() then
					if core.numPartyMembers > 0 and core.quests[bValue].complete[core.playerName] and core.quests[bValue].id and core.quests[bValue].sharable and prdb.shareQuest then
						local wasSelected = GetQuestLogSelection()
						SelectQuestLogEntry(core.quests[bValue].id)
						QuestLogPushQuest()
						SelectQuestLogEntry(wasSelected)
					end
				elseif IsShiftKeyDown() then
					if prdb.trackQuest then
						module:toggleTrackedQuest(nil, bValue)
					end
				elseif IsAltKeyDown() then
					if core.quests[bValue].complete[core.playerName] and core.quests[bValue].id and prdb.abandonQuest then
						local wasSelected = GetQuestLogSelection()
						SelectQuestLogEntry(core.quests[bValue].id)
						SetAbandonQuest()
						local items = GetAbandonQuestItems()
						if items then
							StaticPopup_Hide("ABANDON_QUEST")
							StaticPopup_Show("ABANDON_QUEST_WITH_ITEMS", GetAbandonQuestName(), items)
						else
							StaticPopup_Hide("ABANDON_QUEST_WITH_ITEMS")
							StaticPopup_Show("ABANDON_QUEST", GetAbandonQuestName())
						end
						SelectQuestLogEntry(wasSelected)
					end
				else
					core:displayQuestInformations(bValue)
				end
			elseif button == "MiddleButton" then
				if bValue2 ~= nil then
					core.cbh:Fire("toggleMap", bValue, bValue2)
					if Cartographer_QuestInfo then
						local q = Cartographer_QuestInfo:GetQuest(bValue)
						if not q then return end
						if IsControlKeyDown() and prdb.cqiObjectiveStartNPC then
							if q.start_npc then
								data.quest = q.title_full
								data.obj = nil
								data.npc = q.start_npc
								Cartographer_QuestInfo:OnButtonClick(nil, "start", "LeftButton", data)
							end
						elseif IsShiftKeyDown() and prdb.cqiObjectiveEndNPC then
							if q.end_npc then
								data.quest = q.title_full
								data.obj = nil
								data.npc = q.end_npc
								Cartographer_QuestInfo:OnButtonClick(nil, "end", "LeftButton", data)
							end
						elseif prdb.cqiObjective then
							for _, obj in pairs(q.objs) do
								if obj.title:find(core.questsObjectives[bValue][bValue2].desc, 1, true) then
									data.quest = q.title_full
									data.obj = obj
									data.npc = nil
								end
							end
							if data.obj ~= nil then
								Cartographer_QuestInfo:OnButtonClick(nil, "obj", "LeftButton", data)
							end
						end
					elseif MobMapDotParentFrame and MobMap_ParseQuestObjective and prdb.mmObjective then
						MobMap_ParseQuestObjective(core.questsObjectives[bValue][bValue2].desc)
					end
				end
			end
		end
	end)
	self:SetScript("OnEnter", function()
		if prdb.overTooltip and bType == "quest" then
			if bValue2 ~= nil and core.questsObjectives[bValue] and core.questsObjectives[bValue][bValue2].qType == "object" and core.questsItems[core.questsObjectives[bValue][bValue2].desc].mobDrop then
				GameTooltip:SetOwner(tooltip, anchor)
				GameTooltip:AddLine(LQTr["|cffffffffDrop from:|r"])
				for mobName in pairs(core.questsItems[core.questsObjectives[bValue][bValue2].desc].mobDrop) do
					local zones = ""
					for zone in pairs(core.questsMobsItems[mobName].mobLoc) do
						zones = zones == "" and zone or format("%s, %s", zones, zone)
					end
					if core.questsMobsItems[mobName].items[core.questsObjectives[bValue][bValue2].desc].dropRate then
						GameTooltip:AddDoubleLine(format(LQTr["  |cffffffff%s in|r |cff00ff00%s|r"], mobName, zones), format("|cffffffff%s%%|r", core.questsMobsItems[mobName].items[core.questsObjectives[bValue][bValue2].desc].dropRate))
					else
						GameTooltip:AddLine(format(LQTr["  |cffffffff%s in|r |cff00ff00%s|r"], mobName, zones))
					end
				end
				GameTooltip:Show()
			elseif core.quests[bValue] and core.quests[bValue].id then
				local link = GetQuestLink(core.quests[bValue].id)
				if link then
					GameTooltip:SetOwner(tooltip, anchor)
					GameTooltip:SetHyperlink(link)
				end
			end
		end
	end)
	self:SetScript("OnLeave", function()
		GameTooltip:Hide()
	end)

	indent = indent or 0
	local width, height = basecCellPrototype.SetupCell(self, tooltip, value, justification, font)
	local fs = self.fontString
	fs:SetPoint("TOPLEFT", self, "TOPLEFT", indent, 0)
	fs:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", 0, 0)
	width = width + indent
	if width > maxWidth then
		fs:SetWidth(maxWidth)
		height = fs:GetHeight()
		width = maxWidth
	end
	fs:Show()
	return width, height
end

function cPrototype:ReleaseCell() 
	self:EnableMouse(false) 
	self:ClearAllPoints()
	self:SetScript("OnEnter", nil) 
	self:SetScript("OnLeave", nil) 
	self:SetScript("OnMouseDown", nil) 
end

function module:updateQuestTracker()
	if     (not prdb.toggleTracker) or (numTrackedQuests == 0) or (prdb.hideCompletedQuest == 2 and numTrackedQuests == core.numPartyCompletedTrackedQuests)
		or (prdb.hideCompletedQuest == 1 and numTrackedQuests == core.numPlayerCompletedTrackedQuests)
		or (prdb.onlyCurrentZone and (   not core.questsByZone[currentZone]
									  or core.questsByZone[currentZone].numTrackedQuests == 0
									  or (prdb.hideCompletedQuest == 2 and core.questsByZone[currentZone].numTrackedQuests == core.questsByZone[currentZone].partyQuestsComplete)
									  or (prdb.hideCompletedQuest == 1 and core.questsByZone[currentZone].numTrackedQuests == core.questsByZone[currentZone].playerQuestsComplete)
									  or (prdb.limitQuestLevel and core.questsByZone[currentZone].numTrackedQuests == core.questsByZone[currentZone].questsBML)
									  ))
		or (prdb.limitQuestLevel and numTrackedQuests == numTrackedQuestsBML)
		or (prdb.hideTrackerInCombat and UnitAffectingCombat("player")) then return end
	tracker = LibQTip:Acquire("tomQuest2Tracker", 2, "LEFT", "RIGHT")
	tracker:EnableMouse(true)
	tracker:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
	tracker:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
	tracker:SetScale(prdb.scale)
	tracker:SetFrameStrata("BACKGROUND")
	local tempLine
	if prdb.displayHeader then
		for zone, zData in pairs(core.questsByZone) do
			if (prdb.onlyCurrentZone and zone == currentZone) or not prdb.onlyCurrentZone then
				if (prdb.hideCompletedQuest == 3 and zData.numTrackedQuests > 0) or (prdb.hideCompletedQuest == 2 and zData.numTrackedQuests > zData.partyQuestsComplete) or (prdb.hideCompletedQuest == 1 and zData.numTrackedQuests > zData.playerQuestsComplete) then
					chdb.zoneDisplay[zone] = chdb.zoneDisplay[zone] or 1
					tempLine = tracker:AddLine()
					tracker:SetCell(tempLine, 1, format("|TInterface\\Buttons\\UI-%sButton-Up:%d|t |cff%s%s|r |cff%s%s|r", chdb.zoneDisplay[zone] == 0 and "Plus" or "Minus", prdb.collapseButtonSize, prdb.colorizeHeaderName and zData.color or "ffffff", zone, prdb.colorizeHeaderLevel and zData.color or "ffffff", prdb.hideZoneLevelRange and "" or zData.level ), hTrFont, "LEFT", 2, cProvider, "zone", zone, nil, prdb.maxWidth, 0)
					if chdb.zoneDisplay[zone] == 1 then
						for uid in pairs(zData.quests) do
							if chdb.questsTracked[uid] and ((prdb.hideCompletedQuest == 3) or (prdb.hideCompletedQuest == 2 and core.quests[uid].partyComplete == 0) or (prdb.hideCompletedQuest == 1 and core.quests[uid].playerComplete == 0)) then
								tempLine = tracker:AddLine()
								tracker:SetCell(tempLine, 1, format("|cff%s%s|r|cff%s%s|r %s", prdb.colorizeQuestLevel and core.quests[uid].color or "ffffff", core.quests[uid].tag, prdb.colorizeQuestName and core.quests[uid].color or "ffffff", core.quests[uid].title, core.quests[uid].playersOnQuest > 1 and format("<%d>", core.quests[uid].playersOnQuest) or ""), qTrFont, "LEFT", 1, cProvider, "quest", uid, nil, prdb.maxWidth, prdb.questsCellPadding, trackerAnchor)
								tracker:SetCell(tempLine, 2, core.quests[uid].pqStatus, qTrFont, "RIGHT", 1, cProvider, "quest", uid, nil, prdb.maxWidth, 0, trackerAnchor)
								if core.questsObjectives[uid] then
									for objective, qoData in pairs(core.questsObjectives[uid]) do
										if prdb.hideCompletedObjective == 3 or (prdb.hideCompletedObjective == 2 and qoData.partyComplete == 0) or (prdb.hideCompletedObjective == 1 and qoData.playerComplete == 0) then
											tempLine = tracker:AddLine()
											tracker:SetCell(tempLine, 1, format("|cff%s%s|r", prdb.colorizeQuestObjective and qoData.qoColor or "ffffff", qoData.desc), qoTrFont, "LEFT", 1, cProvider, "quest", uid, objective, prdb.maxWidth, prdb.objectivesCellPadding, trackerAnchor)
											tracker:SetCell(tempLine, 2, prdb.colorizeQuestObjectiveCounter and qoData.qoStatus or qoData.qoStatusNC, qoTrFont, "RIGHT", 1, cProvider, "quest", uid, objective, prdb.maxWidth, 0, trackerAnchor)
										end
									end
								else
									tempLine = tracker:AddLine()
									tracker:SetCell(tempLine, 1, core.quests[uid].objective, qoTrFont, "LEFT", 2, cProvider,"quest", nil, uid, prdb.maxWidth, prdb.objectivesCellPadding, trackerAnchor)
								end
							end
						end
					end
				end
			end
		end
	else
		for uid in pairs(chdb.questsTracked) do
			if uid and core.quests[uid] then
				if (prdb.onlyCurrentZone and core.quests[uid].zone == currentZone) or not prdb.onlyCurrentZone then
					if prdb.hideCompletedQuest == 3 or (prdb.hideCompletedQuest == 2 and not core.quests[uid].partyComplete) or (prdb.hideCompletedQuest == 1 and not core.quests[uid].playerComplete) then
						tempLine = tracker:AddLine()
						tracker:SetCell(tempLine, 1, format("|cff%s%s|r|cff%s%s|r %s", prdb.colorizeQuestLevel and core.quests[uid].color or "ffffff", core.quests[uid].tag, prdb.colorizeQuestName and core.quests[uid].color or "ffffff", core.quests[uid].title, core.quests[uid].playersOnQuest > 1 and format("<%d>", core.quests[uid].playersOnQuest) or ""), qTrFont, "LEFT", 1, cProvider, "quest", uid, nil, prdb.maxWidth, 0, trackerAnchor)
						tracker:SetCell(tempLine, 2, core.quests[uid].pqStatus, qTrFont, "RIGHT", 1, cProvider,  "quests",uid, nil, prdb.maxWidth, 0, trackerAnchor)
						if core.questsObjectives[uid] then
							for objective, qoData in pairs(core.questsObjectives[uid]) do
								if prdb.hideCompletedObjective == 3 or (prdb.hideCompletedObjective == 2 and qoData.partyComplete == 0) or (prdb.hideCompletedObjective == 1 and qoData.playerComplete == 0) then
									tempLine = tracker:AddLine()
									tracker:SetCell(tempLine, 1, format("|cff%s%s|r", prdb.colorizeQuestObjective and qoData.qoColor or "ffffff", qoData.desc), qoTrFont, "LEFT", 1, cProvider, "quest", uid, objective, prdb.maxWidth, prdb.objectivesCellPadding, trackerAnchor)
									tracker:SetCell(tempLine, 2, prdb.colorizeQuestObjectiveCounter and qoData.qoStatus or qoData.qoStatusNC, qoTrFont, "RIGHT", 1, cProvider, "quest", uid, objective, prdb.maxWidth, 0, trackerAnchor)
								end
							end
						else
							tempLine = tracker:AddLine()
							tracker:SetCell(tempLine, 1, core.quests[uid].objective, qoTrFont, "LEFT", 2, cProvider, "quest", uid, nil, prdb.maxWidth, prdb.objectivesCellPadding, trackerAnchor)
						end
					end
				end
			end
		end
	end
	tracker:SetClampedToScreen(true)
	tracker:ClearAllPoints()
	if prdb.anchorPosition == 1 then
		if prdb.showAnchor then
			tracker:SetPoint("TOPLEFT", anchor, "BOTTOMLEFT")
		else
			tracker:SetPoint("TOPLEFT", anchor, "BOTTOMLEFT", 0, 24)
		end
	elseif prdb.anchorPosition == 2 then
		if prdb.showAnchor then
			tracker:SetPoint("TOP", anchor, "BOTTOM")
		else
			tracker:SetPoint("TOP", anchor, "BOTTOM", 0, 24)
		end
	elseif prdb.anchorPosition == 3 then
		if prdb.showAnchor then
			tracker:SetPoint("TOPRIGHT", anchor, "BOTTOMRIGHT")
		else
			tracker:SetPoint("TOPRIGHT", anchor, "BOTTOMRIGHT", 0, 24)
		end
	elseif prdb.anchorPosition == 4 then
		if prdb.showAnchor then
			tracker:SetPoint("BOTTOMLEFT", anchor, "TOPLEFT")
		else
			tracker:SetPoint("BOTTOMLEFT", anchor, "TOPLEFT", 0, -24)
		end
	elseif prdb.anchorPosition == 5 then
		if prdb.showAnchor then
			tracker:SetPoint("BOTTOM", anchor, "TOP")
		else
			tracker:SetPoint("BOTTOM", anchor, "TOP", 0, -24)
		end
	elseif prdb.anchorPosition == 6 then
		if prdb.showAnchor then
			tracker:SetPoint("BOTTOMRIGHT", anchor, "TOPRIGHT")
		else
			tracker:SetPoint("BOTTOMRIGHT", anchor, "TOPRIGHT", 0, -24)
		end
	end
	tracker:Show()
end